#include <string.h>
#include "mymalloc.h"
#include "LibParse.h"
#include "LibEnumMap.h"
#include "LibGroup.h"

void
LibGroup::removeGroup(LibGroup* grp)
{
  d_groupEList.remove(grp);

  LibNameIter iter = grp->nameIter();
  const UString& t = grp->type();
  while (iter) {
    d_groupHash.remove(LibTypedName(t, (iter++).object()));
  }
}

LibGroup::LibGroup(const UString& nm, const UString& t, uint_t ID,
		   LibGroup* parent, const LibFileInfo& f)
  : d_type(t), d_typeID(ID), d_owner(parent), d_info(f)
{
  if (nm.str()) {
    d_nameHashList.append(nm);
  }
}

LibGroup::~LibGroup()
{
   LibAttrIter aiter = attrIter();
   LibAttr* a;
   while ((a = aiter++)) {
     removeAttr(a);
     delete a;
   }
   LibDefineIter diter = defineIter();
   LibDefine* d;
   while ((d = diter++)) {
     removeDefine(d);
     delete d;
   }
   LibGroupIter giter = groupIter();
   LibGroup* g;
   while ((g = giter++)) {
     removeGroup(g);
     delete g;
   }
}

LibGroup*
LibGroup::createGroup(const UString& nm, const UString& typ,
		      const LibFileInfo& f, LibErrorT& err)
{
  if (d_owner) {
    return createGroupInGroup(nm, typ, f, err);
  } else {
    return createTopLevelGroup(nm, typ, f, err);
  }
}

LibGroup*
LibGroup::createTopLevelGroup(const UString& name, const UString& group_type,
			      const LibFileInfo& f, LibErrorT& err)
{
  const libGroupMap *lgm;
  LibGroup *g;
  LibTypedName typedName(group_type, name);

  if (name.str() && d_groupHash.find(typedName)) {
    err = LIB_OBJECT_ALREADY_EXISTS;
    LibParse::debug (LIB_SEVERITY_ERR, err, "LibGroup::createTopLevelGroup: ");
    return NULL;
  }
  lgm = Perfect_Hash_group::lookup_group_name(group_type, strlen(group_type));
  g = new IN_CURRENT_POOL LibGroup(name, group_type,
				  (lgm ? lgm->type : (group_enum)0),
				  this, f);
  d_groupEList.append(g);
  if (name.str()) {
    d_groupHash.quickAdd(typedName).value(g);
  }
  err = LIB_NO_ERROR;
  return g;
}

LibAttr*
LibGroup::createAttr(const UString& name, LibAttrType typ,
                     const LibFileInfo& f, LibErrorT& err)
{
  const libAttrMap *lam;
  LibAttr *attr = NULL;
  LibAttrHash::iterator aiter;

  if (name.str()) {
    if (!(aiter = d_attrHash.find(name))) {
      aiter = d_attrHash.quickAdd(name);
    }
  } else {
    err = LIB_INVALID_NAME;
    LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::createAttr:");
    return NULL;
  }

  lam = Perfect_Hash_attr::lookup_attr_name(name, strlen(name));
  attr = new IN_CURRENT_POOL LibAttr(name, typ,
				    (lam ? lam->type : (attr_enum)0),
				    this, f);
  d_attrEList.append(attr);
  aiter.value().append(attr);
  err = LIB_NO_ERROR;
  return attr;
}

LibDefine*
LibGroup::createDefine(const UString& name, const UString& allowed_group_name,
		       LibValType valType, const LibFileInfo& f,
		       LibErrorT& err)
{
  LibDefine *d;

  if (name.str()) {
    int found;
    
    d = (LibDefine*)d_defineEHash.find(name);
    if (d) {
      char buf[1000], *p, *q;
      strncpy(buf, d->groupType(),999);
      found = 0;
      q = buf;

      if (d->groupType() == allowed_group_name) {
	found = 1;
      }
      if (!found) {
	while ((p = strchr(q, '|'))) {
	  *p = 0;
	  if (q == allowed_group_name) {
	    found = 1;
	    break;
	  }
	  q = p + 1;
	}
      }
      if (found) {
	err = LIB_OBJECT_ALREADY_EXISTS;
	LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::createDefine:");
	return NULL;
      } else {
	char *x =
	  (char *)my_malloc(strlen(d->groupType()) +
			    strlen(allowed_group_name) + 2);
	strcpy(x, d->groupType());
	strcat(x, "|");
	strcat(x, allowed_group_name);
	d->groupType(x);
	my_free(x);
	err = LIB_NO_ERROR;
	return d;
      }
    }
  } else {
    err = LIB_INVALID_NAME;
    LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::createDefine:");
    return NULL;
  }
  d = new IN_CURRENT_POOL LibDefine (name, allowed_group_name,
				    valType, this, f);
  d_defineEHash.quickAdd(d);
  // add d to master define hash which is used during syntax checks
  // to look for defines; the following adds to the hash blindly
  // assuming there is no conflict in names between defines from
  // two different groups - may need to be revisited
  LibParse::addToDefineHash(d->name(), d);
  err = LIB_NO_ERROR;
  return d;
}

int
LibGroup::dontAddHash(uint_t grpType)
{
  switch ((group_enum)grpType) {
  case LIBERTY_GROUPENUM_ff_bank:
    if (moreThanOneName())
      return 1;
  case LIBERTY_GROUPENUM_internal_power:
  case LIBERTY_GROUPENUM_cell_degradation:
  case LIBERTY_GROUPENUM_cell_fall:
  case LIBERTY_GROUPENUM_cell_rise:
  case LIBERTY_GROUPENUM_fall_constraint:
  case LIBERTY_GROUPENUM_fall_propagation:
  case LIBERTY_GROUPENUM_fall_transition:
  case LIBERTY_GROUPENUM_retaining_fall:
  case LIBERTY_GROUPENUM_retaining_rise:
  case LIBERTY_GROUPENUM_rise_constraint:
  case LIBERTY_GROUPENUM_rise_propagation:
  case LIBERTY_GROUPENUM_rise_transition:
  case LIBERTY_GROUPENUM_em_max_toggle_rate:
  case LIBERTY_GROUPENUM_fall_power:
  case LIBERTY_GROUPENUM_power:
  case LIBERTY_GROUPENUM_rise_power:
  case LIBERTY_GROUPENUM_fall_net_delay:
  case LIBERTY_GROUPENUM_fall_transition_degradation:
  case LIBERTY_GROUPENUM_rise_net_delay:
  case LIBERTY_GROUPENUM_rise_transition_degradation:
  case LIBERTY_GROUPENUM_interconnect_delay:
  case LIBERTY_GROUPENUM_scaled_cell:
    return 1;
  default:
    return 0;
  }
  return 0;
}
  
LibGroup*
LibGroup::createGroupInGroup(const UString& name, const UString& group_type,
			     const LibFileInfo& f, LibErrorT& err)
{
  LibTypedName typedName(group_type, name);
  LibGroup *g;
  const libGroupMap *lgm;
  int dont_addhash = 0;

  if (!group_type.str()) {
    err = LIB_INVALID_VALUE;
    LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::createGroupinGroup:");
    return NULL;
  }
  lgm = Perfect_Hash_group::lookup_group_name(group_type, strlen(group_type));
  if (name.str()) {
    if (d_groupHash.find(typedName)) {
      /* disqualify any references here */
      dont_addhash = dontAddHash(lgm->type);
      if (!dont_addhash) {
	err = LIB_OBJECT_ALREADY_EXISTS;
	LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::createGroupinGroup:");
	return NULL;
      }
    }
  }
  g = new IN_CURRENT_POOL LibGroup(name, group_type,
				  (lgm ? lgm->type : (group_enum)0),
				  this, f);
  d_groupEList.append(g);

  if (name.str() && !dont_addhash) {
    d_groupHash.quickAdd(typedName).value(g);
  }
  err = LIB_NO_ERROR;
  return g;
}

void
LibGroup::addName(const UString& name, LibErrorT& err)
{
  LibTypedName typedName(d_type, name);
  int dont_addhash = 0;

  err = LIB_NO_ERROR;
  if (name.str()) {
    if (d_owner && d_owner->d_groupHash.find(typedName)) {
      /* disqualify any references here */
      dont_addhash = dontAddHash(d_typeID);
      if (!dont_addhash) {
	err = LIB_OBJECT_ALREADY_EXISTS;
	LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::addName:");
      }
    }
  } else {
    err = LIB_INVALID_NAME;
    LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::addName:");
    return;
  }
  if (d_owner && err != LIB_OBJECT_ALREADY_EXISTS && !dont_addhash) {
    d_owner->d_groupHash.quickAdd(typedName).value(this);
  }
  d_nameHashList.append(name);
}

void
LibGroup::deleteName(const UString& name, LibErrorT& err)
{
  LibTypedName typedName(d_type, name);

  if (name.str()) {
    if (d_owner && !d_owner->d_groupHash.find(typedName)) {
      err = LIB_OBJECT_NOT_FOUND;
      LibParse::debug (LIB_SEVERITY_ERR, err, "LibGroup::deleteName:");
      return;
    }
  } else {
    err = LIB_INVALID_NAME;
    LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::deleteName:");
    return;
  }
  d_nameHashList.remove(name);
  if (d_owner) {
    d_owner->d_groupHash.remove(typedName);
  }
  err = LIB_NO_ERROR;
}

LibGroup*
LibGroup::findGroup(const UString& name, const UString& typ, LibErrorT& err)
const
{
  LibGroupHash::iterator iter (0);

  err = LIB_NO_ERROR;
  if (name.str()) {
    if (!(iter = d_groupHash.find(LibTypedName(typ, name)))) {
      err = LIB_OBJECT_NOT_FOUND;
      LibParse::debug(LIB_SEVERITY_NOTE, err, "LibGroup::findGroup:");
      return NULL;
    }
  } else {
    err = LIB_INVALID_NAME;
    LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::findGroup:");
    return NULL;
  }
  return iter.value();
}

LibAttrPIter
LibGroup::findAttr(const UString& name, LibErrorT& err)
const
{
  LibAttrHash::iterator aiter;

  err = LIB_NO_ERROR;
  if (name.str()) {
    aiter = d_attrHash.find(name);  
    if (!aiter) {
      err = LIB_OBJECT_NOT_FOUND;
      LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::findAttr:");
      return NULL;
    } else {
      return aiter.value().begin();
    }
  } else {
    err = LIB_INVALID_NAME;
    LibParse::debug(LIB_SEVERITY_NOTE, err, "LibGroup::findAttr: ");
    return NULL;
  }
}

LibDefine*
LibGroup::findDefine(const UString& name, LibErrorT& err)
const
{
  LibDefine* def;

  err = LIB_NO_ERROR;
  if (name.str()) {
    def = (LibDefine*)d_defineEHash.find(name);
    if (!def) {
      err = LIB_OBJECT_NOT_FOUND;
      LibParse::debug(LIB_SEVERITY_NOTE, err, "LibGroup::findDefine:");
      return NULL;
    } else {
      return def;
    }
  } else {
    err = LIB_INVALID_NAME;
    LibParse::debug(LIB_SEVERITY_ERR, err, "LibGroup::findDefine:");
    return NULL;
  }
}

void
LibGroup::moveBefore(LibGroup* targetGroup)
{
  uassert(this != targetGroup && d_owner == targetGroup->d_owner);

  LibGroupIter iter = targetGroup;
  if ((LibGroup*)(--iter) == this) return;
  d_owner->d_groupEList.remove(this);
  d_owner->d_groupEList.insertBefore(targetGroup, this);
}

void
LibGroup::moveAfter(LibGroup* targetGroup)
{
  uassert(this != targetGroup && d_owner == targetGroup->d_owner);

  LibGroupIter iter = targetGroup;
  if ((LibGroup*)(++iter) == this) return;
  d_owner->d_groupEList.remove(this);
  d_owner->d_groupEList.insertAfter(targetGroup, this);
}

static int
lib__name_needs_to_be_quoted(const char *name)
{
  const char *s = name;
  if (*s >= '0' && *s <= '9')
    return 1;
  while (*s) {
    if (*s == ' ')
      return 1;
    if ((*s < '0') || (*s > '9' && *s < 'A') || (*s > 'Z' && *s < 'a')
	|| (*s > 'z'))
      return 1;
    s++;
  }
  return 0;
}

void
LibGroup::dump(FILE* f, const char* indent)
{
  char indent2[200];
  const char* str;
  LibAttr* attr;
  LibGroup* g;
  LibDefine* def;
  LibGroupIter giter;
  LibAttrIter aiter;
  LibNameIter niter;
  LibDefineIter diter;
  int first;


  sprintf(indent2, "%s\t", indent);
  first = 1;
  fprintf(f, "%s%s (", indent, (str_t)d_type);

  // print the names
  niter = nameIter();
  while ( niter) {
    str = *niter++;
    if (!first)
      fprintf(f, ",");
    if (lib__name_needs_to_be_quoted(str))
      fprintf(f, "\"%s\"", str);
    else
      fprintf(f, "%s", str);
    first = 0;
  }
  fprintf(f, ") {\n");

  // print the defines
  diter = defineIter();
  while ((def = diter++)) {
    def->dump(f, indent2);
  }
  // print the attrs
  aiter = attrIter();
  while ((attr = aiter++)) {
    if (attr->type() == LIBERTY__SIMPLE &&
	(attr->name() == "default_operating_conditions" ||
	 attr->name() == "default_wire_load" ||
	 attr->name() == "default_wire_load_selection"))
      continue;
    attr->dump(f, indent2);
  }
  // print the groups
  giter = groupIter();
  while ((g = giter++)) {
    g->dump(f, indent2);
  }
  /* print the attrs */
  aiter = attrIter();
  while ((attr = aiter++)) {
    if (attr->type() == LIBERTY__SIMPLE &&
	attr->name() != "default_operating_conditions" &&
	attr->name() != "default_wire_load" &&
	attr->name() != "default_wire_load_selection")
      continue;
    attr->dump(f, indent2);
  }
  fprintf(f, "%s}\n", indent);
}

void
LibGroup::createFloatingDefineForCellArea(const char* string)
{
  LibFileInfo f;
  LibDefine *d;

  d = new IN_CURRENT_POOL LibDefine(string, "cell", LIBERTY__VAL_DOUBLE,
				   NULL, f);
  d_defineEHash.quickAdd(d);
  LibParse::addToDefineHash(d->name(), d);
}
